/*
 * Copyright (c) 2013,2018 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Wed Jan 16 2013
 */
#include "acrbulkit.h"

#include <sys/mman.h>
#include <sys/param.h>
#include <sys/queue.h>
#include <sys/select.h>
#include <sys/uio.h>
#include <sys/un.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>

#include <assert.h>
#include <ctype.h>
#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <fnmatch.h>
#include <grp.h>
#include <libgen.h>
#include <pwd.h>
#include <regex.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <time.h>
#include <unistd.h>


/* --- gobal data ---------------------------------------------------------- */

char config[MAXPATHLEN] = PACKAGE_PREFIX "/etc/rbulk-dispatchd.conf";
char bind_dir[MAXPATHLEN] = "/var/run";
char rbd_logstatus[MAXPATHLEN] = "";
char rbd_status[MAXPATHLEN] = "";
char rbd_tmpstatus[MAXPATHLEN] = "";
char *run_user = NULL;
char *run_group = NULL;
int grace = 600;
int histlen = 7;
int daemonize = 1;
int daemonized = 0;
int leave = 0;
int child = 0;
int expire = 0;
int proccounter = 0;

LIST_HEAD(, rbd_job) jobs = LIST_HEAD_INITIALIZER(jobs);
LIST_HEAD(, rbd_group) groups = LIST_HEAD_INITIALIZER(groups);
int ngroups = 0, nslots = 0;

uid_t ruid;
gid_t rgid;
char bind_addr[MAXPATHLEN];
sigset_t dfltmask;
int maxwatched = -1;
fd_set watchedset;
struct rbd_proc **fdproc = NULL;
struct rbd_clnt **fdclnt = NULL;

const char *rbd_spawn_str[] = { "init", "exec", "fini", "kill" };
enum rbd_spawn_kind { SPAWN_INIT, SPAWN_EXEC, SPAWN_FINI, SPAWN_KILL };

struct rbd_clnt {
  int fd;
  struct rbd_proc *lastout;

  LIST_HEAD(, rbd_proc) proc;
};

struct rbd_tgt {
  LIST_ENTRY(rbd_tgt) list;

  char key; /* 'g' for groups or 's' for slots */
  const char *pattern;
  int prio;
};

struct rbd_job {
  LIST_ENTRY(rbd_job) list;

  const char *pattern;
  regex_t reg;

  int cost;
  const char *desc;
  const char *exec[SPAWN_KILL+1];

  LIST_HEAD(, rbd_tgt) tgt;
};

struct rbd_proc {
  LIST_ENTRY(rbd_proc) wlist;
  LIST_ENTRY(rbd_proc) clist;

  struct rbd_clnt *clnt;
  struct rbd_slot *slot;
  struct rbd_job *job;
  int prio;
  regmatch_t pmatch[10];

  enum rbd_spawn_kind step;
  char *name;
  char *cmd;
  pid_t pid;
  int progress, status;
  char logbase[256], log[MAXPATHLEN];

  char msg[8192], *rd, *wr;
  size_t mlen;
  time_t bts, ets;
  int fd;
};

struct rbd_slot {
  LIST_ENTRY(rbd_slot) list;

  const char *name;
  const char *desc;
  struct rbd_group *group;
  char *pfix;

  struct rbd_proc *run;
  struct rbd_proc *kill;
  LIST_HEAD(, rbd_proc) wait;
  LIST_HEAD(, rbd_proc) hist;
  time_t ts;
};

struct rbd_group {
  LIST_ENTRY(rbd_group) list;

  const char *name;
  const char *desc;
  int nslots;
  int parallel;
  int running;

  LIST_HEAD(, rbd_slot) slots;
};


/* --- usage --------------------------------------------------------------- */

/** Print usage on a channel.
 *
 * \param[in] channel   Output channel
 * \param[in] argv0     Program invocation name
 */

static void
usage(FILE *channel, char *argv0)
{
  fprintf(
    channel,
    "usage:  %s [-h] [-f] [-l user] [-g group] [-c file] [-b path]\n"
    "                        [-o len] [-d delay] [-s file]\n",
    basename(argv0));
}


/* --- xstuff -------------------------------------------------------------- */

static void
xdelay(int delay)
{
  struct itimerval it = {
    .it_value = { .tv_sec = delay, .tv_usec = 0 },
    .it_interval = { .tv_sec = 0, .tv_usec = 0 }
  };

  (void)setitimer(ITIMER_REAL, &it, NULL);
}

static inline void *
xmalloc(size_t s)
{
  void *p = malloc(s);
  assert(p);
  return p;
}

static inline void *
xrealloc(void *r, size_t s)
{
  void *p = realloc(r, s);
  assert(p);
  return p;
}

static inline char *
xstrdup(char *s)
{
  void *d = strdup(s);
  assert(d);
  return d;
}

static inline struct rbd_proc *
xmallocproc(char *name, regmatch_t pmatch[10], int prio,
            struct rbd_clnt *clnt, struct rbd_slot *slot, struct rbd_job *job,
            int delay)
{
  struct rbd_proc *p = xmalloc(sizeof *p);
  int i;

  p->clnt = clnt;
  p->slot = slot;
  p->job = job;
  p->prio = prio;
  if (*job->pattern == '/') {
    for (i = 0; i < 10; i++) p->pmatch[i] = pmatch[i];
  } else {
    for (i = 0; i < 10; i++) p->pmatch[i].rm_eo = p->pmatch[i].rm_so = -1;
  }

  p->step = SPAWN_INIT;
  p->name = xstrdup(name);
  p->cmd = NULL;
  p->pid = -1;
  p->progress = -1;
  p->status = 0;
  snprintf(p->logbase, sizeof(p->logbase),
           "rbulk-dispatchd/log-%d.txt", ++proccounter);
  snprintf(p->log, sizeof(p->log), rbd_logstatus, p->logbase);
  if (*p->log) unlink(p->log);

  p->wr = p->rd = p->msg;
  p->mlen = sizeof(p->msg);
  memset(p->msg, 0, sizeof(p->mlen));
  p->ets = time(NULL);
  p->bts = p->ets + delay;
  p->fd = -1;

  if (delay) xdelay(delay);

  return p;
}

static inline void
xfreeproc(struct rbd_proc *proc)
{
  if (*proc->log) unlink(proc->log);
  free(proc->name);
  free(proc->cmd);
  free(proc);
}


/* --- log ----------------------------------------------------------------- */

static void
xlog(int prio, const char *fmt, ...)
{
  static int open = 0;
  va_list va;

  if (daemonize) {
    if (!open) {
      openlog("rbulk-distpatchd", LOG_CONS|LOG_PID, LOG_USER);
      open = 1;
    }
    va_start(va, fmt);
    vsyslog(prio, fmt, va);
    va_end(va);
  }
  if (!daemonize || (!daemonized && prio < LOG_INFO)) {
    va_start(va, fmt);
    vwarnx(fmt, va);
    va_end(va);
    fflush(stderr);
  }

  if (prio == LOG_ERR) {
    xlog(LOG_NOTICE, "terminating on fatal error");
    if (open) closelog();
    exit(2);
  }
}


/* --- sighandlers --------------------------------------------------------- */

static void
alarm_sighandler(int sig)
{
  expire = 1;
}

static void
child_sighandler(int sig)
{
  child = 1;
}

static void
term_sighandler(int sig)
{
  leave = sig;
}


/* --- load_statusdata ----------------------------------------------------- */

static void *cssdata, *bodydata, *statedata;
static size_t csslen, bodylen, bodycnt, statelen;

static void
load_statusdata(int unload)
{
#define DATA_PREFIX PACKAGE_PREFIX "/share/rbulkit/html/"
  static char state[256];
  static struct {
    const char *path; void **data; size_t *len; int html;
  } data[] = {
    { DATA_PREFIX "rbulk.css", &cssdata, &csslen, 1 },
    { DATA_PREFIX "rbulk-dispatchd.html", &bodydata, &bodylen, 1 },
    { state, &statedata, &statelen, 0 },
    { NULL, NULL, NULL }
  }, *p;
#undef DATA_PREFIX

  char *s;
  int fd;

  snprintf(state, sizeof state, "%s/rbulk-dispatchd.save", bind_dir);

  for(p = data; p->path; p++) {
    if (p->html && !*rbd_status) continue;
    if (unload) { munmap(*(p->data), *(p->len)); continue; }

    fd = open(p->path, O_RDONLY, 0);
    if (fd < 0) {
      if (p->html)
        xlog(LOG_ERR, "%s: %s", p->path, strerror(errno));
      *(p->data) = NULL;
      continue;
    }

    *(p->len) = lseek(fd, 0, SEEK_END);
    *(p->data) = mmap(NULL, *(p->len),
                      PROT_READ|PROT_WRITE, MAP_FILE|MAP_PRIVATE, fd, 0);
    if (*(p->data) == MAP_FAILED)
      xlog(LOG_ERR, "%s: %s", p->path, strerror(errno));
    close(fd);
  }
  if (unload) return;

  if (statedata) {
    xlog(LOG_NOTICE, "reloaded state from %s", state);
    unlink(state);
    ((char *)statedata)[statelen-1] = *"";
  }

  /* prepare body data for @-style replacements in write_status() */
  if (*rbd_status) {
    for (s = bodydata, bodycnt = 0; strsep(&s, "@"); bodycnt++)
      ;
  }
}


/* --- write_status -------------------------------------------------------- */

static void
write_status(void)
{
  struct rbd_group *grp;
  struct rbd_proc *proc;
  struct rbd_slot *slt;
  size_t l;
  FILE *f;
  char *s;
  int fd;
  int i, n, m;

  if (!*rbd_status) return;
  fd = open(rbd_tmpstatus, O_WRONLY|O_CREAT|O_TRUNC, 0444);
  if (fd < 0) {
    xlog(LOG_WARNING, "open %s: %s", rbd_tmpstatus, strerror(errno));
    return;
  }
  f = fdopen(fd, "w");
  if (!f) {
    xlog(LOG_WARNING, "fdopen %s: %s", rbd_tmpstatus, strerror(errno));
    close(fd);
    return;
  }

  /* @-style replacements: even chunks are raw data, odd chunks are substs */
  for(i=0,s=bodydata; i<bodycnt; i++,s+=l+1) {
    l = strlen(s);
    if (i%2 == 0) { fwrite(s, l, 1, f); continue; }

    if (!strcmp(s, "css"))
      fwrite(cssdata, csslen, 1, f);
    else if (!strcmp(s, "ts"))
      fprintf(f, "%s", ctime(&(time_t){time(NULL)}));
    else if (!strcmp(s, "status")) {
      if (LIST_EMPTY(&groups)) {
        fprintf(f, "<h2>Daemon stopped</h2>\n");
        continue;
      }

      /* header */
      fprintf(
        f, "<table class=dispatch><thead class=dispatch><tr>"
        "<th class=empty>Status</th><th colspan=%d>Slots</th></tr>"
        "<tr><th class=empty></th>\n", nslots);
      LIST_FOREACH(grp, &groups, list) {
        fprintf(f, "<th colspan=%d>%s</th>",
                grp->nslots, grp->desc?grp->desc:grp->name);
      }
      fprintf(f, "</tr><tr><th class=empty></th>");
      LIST_FOREACH(grp, &groups, list) LIST_FOREACH(slt, &grp->slots, list) {
        fprintf(f, "<th>%s</th>", slt->desc?slt->desc:slt->name);
      }

      /* queued */
      fprintf(f, "</tr></thead><tbody><tr class=bdesc>"
              "<td class=desc>Queued</td>");
      LIST_FOREACH(grp, &groups, list) LIST_FOREACH(slt, &grp->slots, list) {
        fprintf(f, "<td class=\"dispatch bdesc\">");
        /* must reverse the list. sigh */
        n = 0; LIST_FOREACH(proc, &slt->wait, wlist) n++;
        if (n > 0) {
          struct rbd_proc *wlist[n];
          m = n; LIST_FOREACH(proc, &slt->wait, wlist) wlist[--m] = proc;
          for(m = 0; m < n; m++) {
            fprintf(
              f,
              "<ul class=job title=\"%s\">"
              "<li class=name>%s</li>"
              "<li class=tag style=\"background:#%02x%02x66\">prio #%d</li>"
              "<li class=\"ts clear\">%.16s</li>" "</ul>",
              wlist[m]->job->desc?wlist[m]->job->desc:"(no description)",
              wlist[m]->name,
              (int)(wlist[m]->prio>127?127+128*(255-wlist[m]->prio)/128.:255),
              (int)(wlist[m]->prio>127?255:127+128*wlist[m]->prio/128.),
              wlist[m]->prio, ctime(&wlist[m]->bts)+4);
          }
        }
        fprintf(f, "</td>");
      }

      /* running */
      fprintf(f, "</tr></tbody><tbody>"
              "<tr class=run><td class=desc>Running</td>");
      LIST_FOREACH(grp, &groups, list) LIST_FOREACH(slt, &grp->slots, list) {
        fprintf(f, "<td class=dispatch>");
        if (slt->run) {
          char status[128], log[512];
          struct stat sb;

          if (slt->kill) {
            snprintf(status, sizeof(status), "<li class=kill>stop</li>");
          } else if (slt->run->step == SPAWN_INIT ||
                     slt->run->step == SPAWN_FINI) {
            if (slt->run->progress >= 0)
              snprintf(status, sizeof(status), "<li class=progress>%s %d%%</li>",
                       rbd_spawn_str[slt->run->step], slt->run->progress);
            else
              snprintf(status, sizeof(status), "<li class=progress>%s</li>",
                       rbd_spawn_str[slt->run->step]);
          } else if (slt->run->progress >= 0) {
            snprintf(status, sizeof(status), "<li class=progress>%d%%</li>",
                     slt->run->progress);
          } else
            *status = *"";

          if (!stat(slt->run->log, &sb))
            snprintf(
              log, sizeof(log), "<a href=\"%s\">Log</a>", slt->run->logbase);
          else
            *log = *"";

          fprintf(
            f,
            "<ul class=job title=\"%s\">"
            "<li class=name>%s</li>"
            "%s<li class=tag style=\"background:#%02x%02x66\">prio #%d</li>"
            "<li class=\"ts clear\">%.16s</li>"
            "<li class=log>%s</li>"
            "<li class=clear></li></ul>",
            slt->run->job->desc?slt->run->job->desc:"(no description)",
            slt->run->name,
            status,
            (int)(slt->run->prio>127?127+128*(255-slt->run->prio)/128.:255),
            (int)(slt->run->prio>127?255:127+128*slt->run->prio/128.),
            slt->run->prio, ctime(&slt->run->bts)+4, log);
        } else
          fprintf(
            f,
            "<ul class=idle>"
            "<li class=ts>%.16s</li><li class=tag>idle</li>"
            "<li class=clear></li></ul>", ctime(&slt->ts)+4);
        fprintf(f, "</td>");
      }

      /* history */
      fprintf(f, "</tr></tbody><tbody><tr class=tdesc>"
              "<td class=desc>Completed</td>");
      LIST_FOREACH(grp, &groups, list) LIST_FOREACH(slt, &grp->slots, list) {
        char ets[26], log[512];
        struct stat sb;

        fprintf(f, "<td class=dispatch>");
        LIST_FOREACH(proc, &slt->hist, wlist) {
          ctime_r(&proc->ets, ets);

          if (!stat(proc->log, &sb))
            snprintf(
              log, sizeof(log), "<a href=\"%s\">Log</a>", proc->logbase);
          else
            *log = *"";

          fprintf(
            f,
            "<ul class=job title=\"%s\">"
            "%s<li class=name>%s</li>"
            "<li class=ts>%.16s</li>" "<li class=\"ts rdesc\">%.16s</li>"
            "<li class=tag style=\"background:#%02x%02x66\">%s %d</li>"
            "<li class=log>%s</li>"
            "<li class=clear></li></ul>",
            proc->job->desc?proc->job->desc:"(no description)",
            proc->step == SPAWN_KILL ?"<li class=kill>kill</li>":"",
            proc->name,
            ctime(&proc->bts)+4, ets+4,
            proc->status?255:136, proc->status?136:255,
            WIFSIGNALED(proc->status)?"sig":"exit",
            WIFSIGNALED(proc->status)?
            WTERMSIG(proc->status):WEXITSTATUS(proc->status),
            log);
        }
        fprintf(f, "</td>");
      }
      fprintf(f, "</tbody></table>\n");
    }
  }

  fclose(f);
  if (rename(rbd_tmpstatus, rbd_status))
    xlog(LOG_WARNING, "%s: %s", rbd_status, strerror(errno));
}


/* --- load_config --------------------------------------------------------- */

#define RBD_WRD	"%255[a-zA-Z0-9_/:-]"
#define RBD_PAT	"%255[][*?a-zA-Z0-9_/%:-]"
#define RBD_REG	"%255[][(|).*?+{,}^$a-zA-Z0-9_/%:-]"

static int
load_config()
{
  enum { CFG_NULL, CFG_GROUP, CFG_JOB } context = CFG_NULL;
  char key[256], value[256], *line, *p, *e;
  struct rbd_group *g;
  struct rbd_slot *s;
  struct rbd_job *j;
  struct rbd_tgt *t;
  int status = 0;
  int lineno = 0;
  int len, rest, i, m;
  FILE *f;

  f = fopen(config, "r");
  if (!f) xlog(LOG_ERR, "%s: %s", config, strerror(errno));

  len = 256;
  p = line = xmalloc(len);

  while (fgets(p, len - (p - line), f) != NULL) {
    ++lineno;
    p = strchr(line, '\n');
    if (p == NULL) {
      if (strlen(line) >= len-1) {
        len *= 2;
        line = xrealloc(line, len);
        p = line + strlen(line);
        continue;
      }

      xlog(LOG_WARNING, "%s:%d: no final newline", config, lineno);
      status = 2;
    } else
      *p = '\0';
    if (p > line && p[-1] == '\\') { p--; continue; }
    if ((p = strchr(line, '#'))) *p = '\0';
    for (p = line; isspace((int)*p); p++);
    if (*p == '\0') goto next;

    /* get section header */
    if (sscanf(p, "[" RBD_WRD " " RBD_REG " %n", key, value, &rest) == 2) {
      if (p[rest]) {
        xlog(LOG_WARNING,
             "%s:%d: extra characters after section", config, lineno);
        status = 2;
      }
      if (value[strlen(value)-1] != ']') {
        xlog(LOG_WARNING, "%s:%d: missing close bracket", config, lineno);
        status = 2;
        goto next;
      }
      value[strlen(value)-1] = *"";

      if (!strcmp(key, "group")) {
        if (sscanf(value, RBD_WRD, key) != 1 || strlen(key) != strlen(value)) {
          xlog(LOG_WARNING, "%s:%d: invalid characters in `%s'",
               config, lineno, value);
          status = 2;
          goto next;
        }

        LIST_FOREACH(g, &groups, list) if (!strcmp(g->name, value))
          xlog(LOG_ERR, "%s:%d: duplicate group %s", config, lineno, g->name);

        g = xmalloc(sizeof(*g));
        g->name = xstrdup(value);
        g->desc = NULL;
        g->nslots = 0;
        g->parallel = 1;
        g->running = 0;
        LIST_INIT(&g->slots);
        LIST_INSERT_HEAD(&groups, g, list);

        ngroups++;
        context = CFG_GROUP;
        goto next;
      } else if (!strcmp(key, "job")) {
        /* check pattern or regexp */
        switch (*value) {
          case '/':
            if (strlen(value) < 3 || value[strlen(value)-1] != '/') {
              xlog(LOG_WARNING, "%s:%d: invalid regexp in `%s'",
                   config, lineno, value);
              status = 2;
              context = CFG_NULL;
              goto next;
            }
            break;

          default:
            if (sscanf(value, RBD_PAT, key) != 1 ||
                strlen(key) != strlen(value)) {
              xlog(LOG_WARNING, "%s:%d: invalid characters in `%s'",
                   config, lineno, value);
              status = 2;
              context = CFG_NULL;
              goto next;
            }
            break;
        }

        LIST_FOREACH(j, &jobs, list) if (!strcmp(j->pattern, value))
          xlog(LOG_ERR, "%s:%d: duplicate job %s", config, lineno, j->pattern);

        j = xmalloc(sizeof(*j));
        j->pattern = xstrdup(value);
        if (*value == '/') {
          value[strlen(value)-1] = *"";
          m = regcomp(&j->reg, value + 1, REG_EXTENDED);
          if (m) {
            regerror(m, &j->reg, value, sizeof(value));
            xlog(LOG_ERR, "%s:%d: %s", config, lineno, key);
            status = 2;
            free(j);
            goto next;
          }
        } else
          regcomp(&j->reg, "^$" /* cannot match anything */, 0);
        j->cost = 1;
        j->desc = NULL;
        for(i = 0; i<= SPAWN_KILL; i++)
          j->exec[i] = NULL;
        LIST_INIT(&j->tgt);
        LIST_INSERT_HEAD(&jobs, j, list);

        context = CFG_JOB;
        goto next;
      }

      xlog(LOG_WARNING, "%s:%d: syntax error", config, lineno);
      status = 2;
      goto next;
    }

    /* parse section contents */
    if (sscanf(p, RBD_WRD " %n", key, &rest) != 1) {
      xlog(LOG_WARNING, "%s:%d: syntax error", config, lineno);
      status = 2;
      continue;
    }
    p += rest;

    switch(context) {
      case CFG_GROUP:
        g = LIST_FIRST(&groups);

        if (!strcmp(key, "desc")) {
          if (g->desc) {
            xlog(LOG_WARNING, "%s:%d: duplicate desc", config, lineno);
            status = 2;
          } else
            g->desc = xstrdup(p);
          goto next;

        } else if (!strcmp(key, "parallel")) {
          g->parallel = strtol(p, &e, 0);
          if (*e || g->parallel < 1 || g->parallel > 1024) {
            xlog(LOG_WARNING, "%s:%d: integer %s invalid or out of range",
                 config, lineno, p);
            status = 2;
          }
          goto next;

        } else if (!strcmp(key, "slot")) {
          if (sscanf(p, RBD_WRD " %n", value, &rest) != 1) {
            xlog(LOG_WARNING, "%s:%d: bad slot name %s", config, lineno, p);
            status = 2;
            goto next;
          }
          p += rest;
          s = xmalloc(sizeof(*s));
          s->name = xstrdup(value);
          s->desc = *p ? xstrdup(p) : NULL;
          s->group = g;
          if (asprintf(&s->pfix, "[%s] + ", s->name));
          s->run = s->kill = NULL;
          s->ts = time(NULL);
          LIST_INIT(&s->wait);
          LIST_INIT(&s->hist);
          LIST_INSERT_HEAD(&g->slots, s, list);
          g->nslots++;
          nslots++;
          goto next;
        }
        break;

      case CFG_JOB:
        j = LIST_FIRST(&jobs);

        if (!strcmp(key, "desc")) {
          if (j->desc) {
            xlog(LOG_WARNING, "%s:%d: duplicate desc", config, lineno);
            status = 2;
          } else
            j->desc = xstrdup(p);
          goto next;

        } else if (!strcmp(key, "cost")) {
          j->cost = strtol(p, &e, 0);
          if (*e || j->cost < 0 || j->cost > 255) {
            xlog(LOG_WARNING, "%s:%d: integer %s invalid or out of range",
                 config, lineno, p);
            status = 2;
          }
          goto next;

        } else if (!strcmp(key, "slot") || !strcmp(key, "group")) {
          if (sscanf(p, RBD_PAT " %n", value, &rest) != 1) {
            xlog(LOG_WARNING, "%s:%d: bad %s name %s", config, lineno, key, p);
            status = 2;
            goto next;
          }
          p += rest;
          if (!strncmp(p, "prio", sizeof("prio")-1) &&
              isspace((int)p[sizeof("prio")-1])) {
            i = strtol(p+sizeof("prio"), &e, 0);
            if (*e || i < 0 || i > 255) {
              xlog(LOG_WARNING, "%s:%d: integer %s invalid or out of range",
                   config, lineno, p+sizeof("prio"));
              status = 2;
              goto next;
            }
          } else if (!*p) {
            i = 255;
          } else {
            xlog(LOG_WARNING, "%s:%d: syntax error at %s", config, lineno, p);
            status = 2;
            goto next;
          }

          t = xmalloc(sizeof(*t));
          t->key = *key;
          t->pattern = xstrdup(value);
          t->prio = i;
          LIST_INSERT_HEAD(&j->tgt, t, list);
          goto next;

        } else {
          for(i = 0; i <= SPAWN_KILL; i++) {
            if (strcmp(key, rbd_spawn_str[i])) continue;

            if (j->exec[i]) {
              xlog(LOG_WARNING, "%s:%d: duplicate %s in job %s",
                   config, lineno, rbd_spawn_str[i], j->pattern);
              status = 2;
              goto next;
            }
            j->exec[i] = xstrdup(p);
            goto next;
          }
        }
        break;

      case CFG_NULL: break;
    }

    xlog(LOG_ERR, "%s:%d: syntax error", config, lineno);
next:
    p = line;
  }

  free(line);
  fclose(f);

  if (!status) {
    LIST_FOREACH(g, &groups, list) {
      xlog(LOG_INFO, "config: [group %s]%s%s",
           g->name, g->desc?" # ":"", g->desc?g->desc:"");
      xlog(LOG_INFO, "config:   parallel %d", g->parallel);
      LIST_FOREACH(s, &g->slots, list) {
        xlog(LOG_INFO, "config:   slot %s%s%s",
             s->name, s->desc?" # ":"", s->desc?s->desc:"");
      }
    }

    LIST_FOREACH(j, &jobs, list) {
      xlog(LOG_INFO, "config: [job %s]%s%s",
           j->pattern, j->desc?" # ":"", j->desc?j->desc:"");
      xlog(LOG_INFO, "config:   cost %d", j->cost);
      LIST_FOREACH(t, &j->tgt, list) {
        xlog(LOG_INFO, "config:   %s %s prio %d",
             t->key == 'g' ? "group" : "slot", t->pattern, t->prio);
      }
      for(i = 0; i <= SPAWN_KILL; i++) {
        if (j->exec[i])
          xlog(LOG_INFO, "config:   %s %s", rbd_spawn_str[i], j->exec[i]);
        else if (i == SPAWN_EXEC) {
          xlog(LOG_INFO, "config:   missing exec");
          status = 2;
        }
      }
    }
  }
  return status;
}


/* --- fdwatch ------------------------------------------------------------- */

static void
fdwatch(int s)
{
  if (s > maxwatched) {
    fdproc = xrealloc(fdproc, (s+1) * sizeof(*fdproc));
    fdclnt = xrealloc(fdclnt, (s+1) * sizeof(*fdclnt));
    for(++maxwatched; maxwatched<s; maxwatched++) {
      fdproc[maxwatched] = NULL;
      fdclnt[maxwatched] = NULL;
    }
    fdproc[s] = NULL;
    fdclnt[s] = NULL;
  }
  FD_SET(s, &watchedset);
}


/* --- setup_socket --------------------------------------------------------- */

static void
setup_socket(const char *bind_dir)
{
  struct sockaddr_un un;
  int s;

  s = socket(PF_LOCAL, SOCK_STREAM|SOCK_CLOEXEC, 0);
  if (s < 0) xlog(LOG_ERR, "socket: %s", strerror(errno));

  memset(&un, 0, sizeof(un));
  un.sun_family = PF_LOCAL;
  strncpy(un.sun_path, bind_dir,
          sizeof(un.sun_path)-sizeof("/rbulk-dispatchd"));
  strcat(un.sun_path, "/rbulk-dispatchd");

  unlink(un.sun_path);
  if (bind(s, (struct sockaddr *)&un, sizeof(un)) < 0)
    xlog(LOG_ERR, "binding on %s: %s", un.sun_path, strerror(errno));
  if (listen(s, 16)) {
    unlink(un.sun_path);
    xlog(LOG_ERR, "listening on %s: %s", un.sun_path, strerror(errno));
  }

  strncpy(bind_addr, un.sun_path, MAXPATHLEN);
  if (chmod(un.sun_path, S_IRUSR|S_IWUSR))
    xlog(LOG_ERR, "chmod on %s: %s", un.sun_path, strerror(errno));

  fdwatch(s);

  xlog(LOG_INFO, "listening on %s", un.sun_path);
}


/* --- join_proc ----------------------------------------------------------- */

static void
join_proc(struct rbd_proc *proc)
{
  struct rbd_slot *slot;
  struct rbd_clnt *clnt;
  struct rbd_proc *p;
  int n;

  slot = proc->slot;

  /* remove from client */
  LIST_REMOVE(proc, clist);
  clnt = proc->clnt;
  if (LIST_EMPTY(&clnt->proc)) {
    xlog(LOG_INFO, "client #%d terminated", clnt->fd);
    FD_CLR(clnt->fd, &watchedset);
    fdclnt[clnt->fd] = NULL;
    if (clnt->fd > STDERR_FILENO) close(clnt->fd);
    free(clnt);
  } else
    if (clnt->lastout == proc) clnt->lastout = NULL;

  /* cleanup */
  if (slot->run == proc) {
    slot->run = NULL;
    slot->group->running -= proc->job->cost;
  } else if (slot->kill == proc) {
    slot->kill = NULL;
  }
  if (proc->fd >= 0) {
    fdproc[proc->fd] = NULL;
    close(proc->fd);
    proc->fd = -1;
  }

  /* put in history */
  proc->ets = slot->ts = time(NULL);
  LIST_INSERT_HEAD(&slot->hist, proc, wlist);

  /* clamp history */
  n = 0;
  LIST_FOREACH(proc, &slot->hist, wlist)
    if (++n >= histlen) {
      while((p = LIST_NEXT(proc, wlist))) {
        LIST_REMOVE(p, wlist);
        xfreeproc(p);
      }
      break;
    }
}


/* --- clntmsg ------------------------------------------------------------- */

static void
clntmsg(struct rbd_clnt *clnt, const char *fmt, ...)
{
  char msg[256];
  ssize_t l, n;
  va_list va;

  if (!FD_ISSET(clnt->fd, &watchedset)) return;

  va_start(va, fmt);
  n = vsnprintf(msg, sizeof(msg), fmt, va);
  if (n >= sizeof(msg)) n = sizeof(msg) -1;
  msg[n++] = '\n';
  va_end(va);

  l = write(clnt->fd, msg, n);
  if (l < 0) {
    xlog(LOG_INFO, "client #%d disconnected: %s", clnt->fd, strerror(errno));
    shutdown(clnt->fd, SHUT_WR);
    FD_CLR(clnt->fd, &watchedset);
    return;
  }
}


/* --- spawn_job ----------------------------------------------------------- */

static void
spawn_job(struct rbd_slot *slot, enum rbd_spawn_kind k)
{
  char *argp[] = { "sh", "-c", NULL, NULL };
  struct rbd_proc *proc;
  const char *p, *q;
  char *c, repl[256];
  pid_t pid;
  int fd[2];
  int len;

  switch(k) {
    case SPAWN_INIT:
      /* pop next job */
      assert(!slot->run && !slot->kill);
      proc = LIST_FIRST(&slot->wait);
      if (!proc) return;
      if (slot->group->running + proc->job->cost > slot->group->parallel)
        return;
      LIST_REMOVE(proc, wlist);

      proc->step = SPAWN_INIT;
      proc->bts = time(NULL);
      p = proc->job->exec[SPAWN_INIT];
      if (!p) {
        k = proc->step = SPAWN_EXEC;
        p = proc->job->exec[SPAWN_EXEC];
      }

      slot->run = proc;
      slot->group->running += proc->job->cost;
      break;

    case SPAWN_EXEC:
    case SPAWN_FINI:
      /* switch init -> exec -> fini */
      assert(slot->run && slot->run->step < k);
      assert(k != SPAWN_EXEC || !slot->kill);

      proc = slot->run;
      proc->step = k;
      p = proc->job->exec[k];
      assert(p);
      break;

    case SPAWN_KILL:
      /* kill running job */
      assert(slot->run && !slot->kill);
      if (slot->run->step != SPAWN_EXEC) {
        slot->run->step = SPAWN_FINI;
        return;
      }

      p = slot->run->job->exec[SPAWN_KILL];
      if (!p) return;

      proc = xmallocproc(slot->run->name, slot->run->pmatch, 0,
                         slot->run->clnt, slot, slot->run->job, 0);
      proc->step = SPAWN_KILL;
      proc->bts = time(NULL);
      LIST_INSERT_HEAD(&proc->clnt->proc, proc, clist);

      slot->kill = proc;
      break;
  }

  free(proc->cmd);
  if (proc->fd >= 0) {
    fdproc[proc->fd] = NULL;
    close(proc->fd);
    proc->fd = -1;
  }

  /* build command - replace %s, %p & %N */
  proc->cmd = xmalloc(len = 64);
  for(c = proc->cmd; *p; p++) {
    if (c - proc->cmd >= len-1) {
      proc->cmd = xrealloc(proc->cmd, len*=2);
      c = proc->cmd + len/2-1;
    }
    if (*p != '%') { *c++ = *p; continue; }
    p++;
    switch(*p) {
      case '\0': p--; q = ""; break;
      case '%': q = "%"; break;
      case 'j': q = proc->name; break;
      case 's': q = slot->name; break;

      case 'p':
        snprintf(repl, sizeof repl, "%d", slot->run->pid); q = repl;
        break;

      case '0': case '1': case '2': case '3': case '4':
      case '5': case '6': case '7': case '8': case '9': {
        regmatch_t *m = proc->pmatch + (*p - '0');
        if (m->rm_so < 0 || m->rm_eo < m->rm_so)
          q = "";
        else {
          snprintf(repl, sizeof repl, "%.*s",
                   (int)(m->rm_eo - m->rm_so), proc->name + m->rm_so);
          q = repl;
        }
        break;
      }

      default: q = ""; break;
    }
    for(; *q; *c++ = *q++) {
      if (c - proc->cmd >= len-1) {
        proc->cmd = xrealloc(proc->cmd, len*=2);
        c = proc->cmd + len/2-1;
      }
    }
  }
  *c = *"";
  argp[2] = proc->cmd;

  /* spawn */
  if (k == SPAWN_KILL) {
    xlog(LOG_INFO, "%scancelling `%s'", slot->pfix, slot->run->cmd);
    clntmsg(slot->run->clnt, "%scancelling `%s'", slot->pfix, slot->run->cmd);
  }

  xlog(LOG_INFO,
       "%sspawning %s `%s'", slot->pfix, rbd_spawn_str[k], proc->cmd);
  clntmsg(proc->clnt,
          "%sspawning %s `%s'", slot->pfix, rbd_spawn_str[k], proc->cmd);
  proc->clnt->lastout = proc;

  if (pipe(fd)) {
    xlog(LOG_WARNING, "%spipe: %s", slot->pfix, strerror(errno));
    clntmsg(proc->clnt, "%spipe: %s", slot->pfix, strerror(errno));
    join_proc(proc);
    return;
  }
  fcntl(fd[0], F_SETFL, O_NONBLOCK);
  fcntl(fd[0], F_SETFD, FD_CLOEXEC);
  proc->fd = fd[0];
  fdwatch(fd[0]);
  fdproc[fd[0]] = proc;

  switch((pid = fork())) {
    case -1:
      xlog(LOG_WARNING, "vfork: %s", strerror(errno));
      clntmsg(proc->clnt, "vfork: %s", strerror(errno));
      close(fd[1]);
      join_proc(proc);
      return;

    case 0: /* child */ {
      uid_t uid = geteuid();
      gid_t gid = getegid();

      if (run_user && seteuid(ruid)) {
        xlog(LOG_WARNING, "seteuid: %s", strerror(errno));
        clntmsg(proc->clnt, "seteuid: %s", strerror(errno));
        _exit(127);
      }
      if (run_group && setgid(gid)) {
        xlog(LOG_WARNING, "setgid: %s", strerror(errno));
        clntmsg(proc->clnt, "setgid: %s", strerror(errno));
        _exit(127);
      }
      if (run_user && setuid(uid)) {
        xlog(LOG_WARNING, "setuid: %s", strerror(errno));
        clntmsg(proc->clnt, "setuid: %s", strerror(errno));
        _exit(127);
      }

      setsid();
      sigprocmask(SIG_SETMASK, &dfltmask, NULL);

      fd[0] = open("/dev/null", O_RDWR, 0);
      if (fd[0] != -1) {
        dup2(fd[0], STDIN_FILENO);
        if (fd[0] != STDIN_FILENO) close(fd[0]);
      }
      dup2(fd[1], STDOUT_FILENO);
      dup2(fd[1], STDERR_FILENO);
      if (fd[1] > STDERR_FILENO) close(fd[1]);

      execvp(argp[0], argp);
      xlog(LOG_WARNING, "execvp: %s", strerror(errno));
      _exit(127);
    }
  }

  proc->pid = pid;
  close(fd[1]);
}


/* --- shuffle_spawn ------------------------------------------------------- */

static void
shuffle_spawn(void)
{
  time_t ts = time(NULL);
  struct rbd_proc *empty[nslots];
  struct rbd_group *g;
  struct rbd_slot *s;
  struct rbd_proc *p;
  int n, i, j, k;

  /* record jobs in empty slots with scheduled jobs ready */
  n = 0;
  LIST_FOREACH(g, &groups, list) {
    LIST_FOREACH(s, &g->slots, list) {
      p = LIST_FIRST(&s->wait);
      if (!s->run && !s->kill && p && p->bts <= ts)
        empty[n++] = p;
    }
  }
  assert(n <= nslots);

  /* shuffle a bit */
  for(k = n-1; k > 0; k--) {
    j = random() % (k+1);
    if (j == k) continue;

    p = empty[j];
    empty[j] = empty[k];
    empty[k] = p;
  }

  /* sort by prio */
  i = n;
  do {
    j = 0;
    for(k = 1; k < i; k++) {
      p = empty[k-1];
      if (p->prio > empty[k]->prio) {
        empty[k-1] = empty[k];
        empty[k] = p;
        j = k;
      }
    }
    i = j;
  } while (i > 0);

  /* spawn */
  for(k = 0; k < n; k++) {
    p = empty[k];
    g = p->slot->group;
    if (g->running + p->job->cost > g->parallel) continue;

    assert(p->slot->run == NULL);
    spawn_job(p->slot, SPAWN_INIT);
  }

  write_status();
}


/* --- send_output --------------------------------------------------------- */

static void
send_output(struct rbd_proc *proc, int flush)
{
  struct iovec iov[4];
  char msg[256];
  ssize_t l;
  void *nl;
  int log;
  int n;

  if (!FD_ISSET(proc->clnt->fd, &watchedset) && !*proc->log) {
    proc->rd = proc->wr;
    return;
  }

  /* if needed, prefix chunk of output with job name */
  iov[0].iov_base = msg;
  if (FD_ISSET(proc->clnt->fd, &watchedset) && proc->clnt->lastout != proc) {
    iov[0].iov_len =
      snprintf(msg, sizeof(msg),
               "%srunning `%s'", proc->slot->pfix, proc->cmd);
    if (iov[0].iov_len >= sizeof(msg)) iov[0].iov_len = sizeof(msg)-1;
    msg[iov[0].iov_len++] = '\n';
  } else
    iov[0].iov_len = 0;

  /* output data */
  iov[1].iov_base = proc->rd;
  iov[2].iov_base = proc->msg;
  if (proc->rd < proc->wr) {
    iov[1].iov_len = proc->wr - proc->rd;
    iov[2].iov_len = 0;
  } else {
    iov[1].iov_len = &proc->msg[proc->mlen] - proc->rd;
    iov[2].iov_len = proc->wr - proc->msg;
  }

  /* if not flushing, print up to last \n only */
  if (!flush) {
    nl = memrchr(iov[2].iov_base, '\n', iov[2].iov_len);
    if (nl) {
      iov[2].iov_len = nl - iov[2].iov_base + 1;
    } else {
      nl = memrchr(iov[1].iov_base, '\n', iov[1].iov_len);
      if (nl) {
        iov[1].iov_len = nl - iov[1].iov_base + 1;
        iov[2].iov_len = 0;
      } else if (proc->rd != proc->wr) {
        /* no \n, not flushing and buffer not full: wait for more data */
        return;
      }
    }
  }

  /* make sure to always end with \n */
  iov[3].iov_base = "\n";
  iov[3].iov_len = 0;
  for(n = 2; n >= 0 && iov[n].iov_len == 0; n--);
  if (((char *)iov[n].iov_base)[iov[n].iov_len-1] != '\n')
    iov[3].iov_len = sizeof("\n")-1;

  /* print */
  if (*proc->log) {
    log = open(proc->log, O_WRONLY|O_APPEND|O_CREAT,
               S_IRUSR|S_IWUSR | S_IRGRP | S_IROTH);
    if (log < 0) {
      xlog(LOG_INFO, "%scannot log to %s for `%s': %s",
           proc->slot->pfix, proc->log, proc->cmd, strerror(errno));
      *proc->log = *"";
    } else {
      l = writev(log, iov+1, 3);
      if (l < 0) {
        xlog(LOG_INFO, "%scannot log to %s for `%s': %s",
             proc->slot->pfix, proc->log, proc->cmd, strerror(errno));
        *proc->log = *"";
      }
      close(log);
    }
  }

  if (FD_ISSET(proc->clnt->fd, &watchedset)) {
    l = writev(proc->clnt->fd, iov, 4);
    if (l < 0) {
      xlog(LOG_INFO, "client #%d disconnected: %s",
           proc->clnt->fd, strerror(errno));
      shutdown(proc->clnt->fd, SHUT_WR);
      FD_CLR(proc->clnt->fd, &watchedset);
      proc->rd = proc->wr;
      return;
    }
    if (l > 0)
      proc->clnt->lastout = proc;
  }

  proc->rd += iov[1].iov_len + iov[2].iov_len;
  if (proc->rd >= &proc->msg[proc->mlen])
    proc->rd -= proc->mlen;
}


/* --- sched_job ----------------------------------------------------------- */

static void
sched_job(struct rbd_clnt *clnt, char *name)
{
  struct rbd_slot *slist[nslots];
  int sprio[nslots];

  struct rbd_proc *proc, *w;
  struct rbd_group *grp;
  struct rbd_slot *slt;
  struct rbd_tgt *tgt;
  struct rbd_job *job;
  regmatch_t pmatch[10];
  char tgtname[256], repl[256];
  const char *p;
  char *t, *q;
  int i, s, n;

  /* look for a job */
  LIST_FOREACH(job, &jobs, list) {
    if (*job->pattern == '/') {
      if (!regexec(&job->reg, name, 10, pmatch, 0)) break;
    } else {
      if (!fnmatch(job->pattern, name, 0)) break;
    }
  }
  if (!job) {
    xlog(LOG_WARNING,
         "client #%d requested unknown job `%s'", clnt->fd, name);
    clntmsg(clnt, "unknown job `%s'", name);
    return;
  }
  xlog(LOG_INFO, "client #%d requested job `%s'", clnt->fd, job->pattern);

  /* build target slots */
  n = 0;
  LIST_FOREACH(tgt, &job->tgt, list) {
    /* compute % substitutions */
    for(t = tgtname, p = tgt->pattern; *p; p++) {
      if (t - tgtname >= sizeof(tgtname) - 1) break;

      if (*p != '%') { *t++ = *p; continue; }
      p++;
      switch(*p) {
        case '\0': p--; q = ""; break;
        case '%': q = "%"; break;
        case 'j': q = name; break;

        case '0': case '1': case '2': case '3': case '4':
        case '5': case '6': case '7': case '8': case '9': {
          regmatch_t *m = pmatch + (*p - '0');
          if (*job->pattern != '/' || m->rm_so < 0 || m->rm_eo < m->rm_so)
            q = "ø"; /* so that this matches nothing */
          else {
            snprintf(repl, sizeof repl, "%.*s",
                     (int)(m->rm_eo - m->rm_so), name + m->rm_so);
            q = repl;
          }
          break;
        }

        default: q = ""; break;
      }
      for(; *q; *t++ = *q++) {
        if (t - tgtname >= sizeof(tgtname) - 1) break;
      }
    }
    *t = *"";

    /* match slots */
    LIST_FOREACH(grp, &groups, list) LIST_FOREACH(slt, &grp->slots, list) {
      if ((tgt->key == 's' && !fnmatch(tgtname, slt->name, 0)) ||
          (tgt->key == 'g' && !fnmatch(tgtname, grp->name, 0))) {
        for(s = 0; s < n; s++) if (slist[s] == slt) break;
        if (s >= n) {
          slist[n] = slt;
          sprio[n] = tgt->prio;
          n++;
        }
      }
    }
  }
  assert(n <= nslots);
  if (!n) {
    xlog(LOG_WARNING, "job `%s' matches no slot", name);
    clntmsg(clnt, "job `%s' matches no slot", name);
    return;
  }

  /* for all target slots */
  for(s = 0; s < n; s++) {
    slt = slist[s];

    /* if job is already scheduled, just update the client */
    i=0;
    LIST_FOREACH(w, &slt->wait, wlist) {
      if (w->job == job) {
        clntmsg(w->clnt, "%scancelled `%s' at #%d", slt->pfix, w->name, i);
        LIST_REMOVE(w, clist);
        if (LIST_EMPTY(&w->clnt->proc)) {
          xlog(LOG_INFO, "client #%d terminated", w->clnt->fd);
          FD_CLR(w->clnt->fd, &watchedset);
          fdclnt[w->clnt->fd] = NULL;
          if (w->clnt->fd > STDERR_FILENO) close(w->clnt->fd);
          free(w->clnt);
        }

        w->clnt = clnt;
        w->bts = time(NULL) + grace;
        xdelay(grace);

        LIST_INSERT_HEAD(&clnt->proc, w, clist);
        xlog(LOG_INFO, "%srescheduled `%s' at #%d", slt->pfix, w->name, i);
        clntmsg(clnt, "%sscheduled `%s' at #%d", slt->pfix, w->name, i);
        clnt->lastout = w;
        break;
      }
      ++i;
    }
    if (w) continue;

    /* enqueue after higher prio targets */
    proc = xmallocproc(name, pmatch, sprio[s], clnt, slt, job, grace);

    i=0;
    if (LIST_EMPTY(&slt->wait))
      LIST_INSERT_HEAD(&slt->wait, proc, wlist);
    else LIST_FOREACH(w, &slt->wait, wlist) {
        if (w->prio > proc->prio) {
          LIST_INSERT_BEFORE(w, proc, wlist);
          break;
        }
        ++i;
        if (!LIST_NEXT(w, wlist)) {
          LIST_INSERT_AFTER(w, proc, wlist);
          break;
        }
      }

    LIST_INSERT_HEAD(&clnt->proc, proc, clist);
    xlog(LOG_INFO, "%sscheduled `%s' at #%d", slt->pfix, proc->name, i);
    clntmsg(clnt, "%sscheduled `%s' at #%d", slt->pfix, proc->name, i);
    clnt->lastout = proc;

    /* if the slot is running a job that has a lower prio or that is the same
     * job as the new one, and it is killable, kill it. Reschedule later if it
     * is a different job.
     */
    if (slt->run && !slt->kill && slt->run->job->exec[SPAWN_KILL]) {
      if (slt->run->step < SPAWN_FINI && slt->run->prio > proc->prio) {
        w = xmallocproc(
          slt->run->name, slt->run->pmatch, slt->run->prio,
          slt->run->clnt, slt, slt->run->job, grace);
        LIST_INSERT_AFTER(proc, w, wlist);
        LIST_INSERT_HEAD(&w->clnt->proc, w, clist);
        xlog(LOG_INFO, "%srescheduled `%s' at #%d", slt->pfix, w->name, i+1);
        clntmsg(w->clnt, "%srescheduled `%s' at #%d",
                slt->pfix, w->name, i+1);
        clnt->lastout = w;
      }

      if (slt->run->prio > proc->prio || slt->run->job == proc->job)
        spawn_job(slt, SPAWN_KILL);
    }
  }
}


/* --- recv_job ------------------------------------------------------------ */

static void
recv_job(struct rbd_clnt *clnt)
{
  char msg[4096], k[256], *p;
  ssize_t l;

  l = read(clnt->fd, msg, sizeof msg);
  if (l < 0) {
    xlog(LOG_WARNING, "client #%d read: %s", clnt->fd, strerror(errno));
    goto eof;
  }
  if (l == 0) {
    xlog(LOG_WARNING, "client #%d disconnected", clnt->fd);
    goto eof;
  }
  if (l >= sizeof msg) {
    xlog(LOG_WARNING, "message too big on client #%d", clnt->fd);
    clntmsg(clnt, "message too big");
    goto eof;
  }
  msg[l] = *"";

  /* lookup jobs */
  for(p = strtok(msg, " \n\t"); p; p = strtok(NULL, " \n\t")) {
    if (sscanf(p, RBD_WRD, k) != 1 || strlen(k) != strlen(p)) {
      xlog(LOG_WARNING,
           "client #%d requested invalid job name `%s'", clnt->fd, p);
      clntmsg(clnt, "invalid job name `%s'", p);
      continue;
    }
    sched_job(clnt, p);
  }

  if (LIST_EMPTY(&clnt->proc)) goto eof;

  /* spawn idle slots - if any */
  shuffle_spawn();
  return;

eof:
  FD_CLR(clnt->fd, &watchedset);
  shutdown(clnt->fd, SHUT_WR);
  if (LIST_EMPTY(&clnt->proc)) {
    xlog(LOG_INFO, "client #%d terminated", clnt->fd);
    fdclnt[clnt->fd] = NULL;
    close(clnt->fd);
    free(clnt);
  }
}


/* --- join_child ---------------------------------------------------------- */

static void
join_child(struct rbd_proc *proc)
{
  struct rbd_clnt *clnt = proc->clnt;
  struct rbd_slot *slot = proc->slot;

  /* print status */
  if (WIFSIGNALED(proc->status))
    xlog(LOG_INFO, "%s%s%s `%s'", slot->pfix,
            strsignal(WTERMSIG(proc->status)),
            WCOREDUMP(proc->status)?" (core dumped)":"", proc->cmd);
  else if (WIFEXITED(proc->status) && WEXITSTATUS(proc->status))
    xlog(LOG_INFO, "%sexit %d `%s'", slot->pfix,
            WEXITSTATUS(proc->status), proc->cmd);
  else
    xlog(LOG_INFO, "%sdone `%s'", slot->pfix, proc->cmd);

  if (WIFSIGNALED(proc->status))
    clntmsg(clnt, "%s%s%s `%s'", slot->pfix,
            strsignal(WTERMSIG(proc->status)),
            WCOREDUMP(proc->status)?" (core dumped)":"", proc->cmd);
  else if (WIFEXITED(proc->status) && WEXITSTATUS(proc->status))
    clntmsg(clnt, "%sexit %d `%s'", slot->pfix,
            WEXITSTATUS(proc->status), proc->cmd);
  else
    clntmsg(clnt, "%sdone `%s'", slot->pfix, proc->cmd);
  clnt->lastout = proc;

  /* update state */
  if (slot->run == proc) {
    switch (proc->step) {
      case SPAWN_INIT:
        if (WIFEXITED(proc->status) && !WEXITSTATUS(proc->status))
          spawn_job(slot, SPAWN_EXEC);
        else
          join_proc(proc);
        break;

      case SPAWN_EXEC:
        if (proc->job->exec[SPAWN_FINI])
          spawn_job(slot, SPAWN_FINI);
        else
          join_proc(proc);
        break;

      case SPAWN_FINI:
      case SPAWN_KILL:
        join_proc(proc);
        break;
    }
  } else
    join_proc(proc);

  if (slot->run == NULL && slot->kill == NULL) {
    /* spawn idle slots - if any */
    shuffle_spawn();
  } else
    write_status();
}


/* --- recv_output --------------------------------------------------------- */

static void
recv_output(struct rbd_proc *proc)
{
  struct iovec iov[2];
  ssize_t l;
  char *t, *d;
  int i, p;

  /* receive in ring buffer */
  iov[0].iov_base = proc->wr;
  iov[1].iov_base = proc->msg;
  if (proc->wr < proc->rd) {
    iov[0].iov_len = proc->rd - proc->wr;
    iov[1].iov_len = 0;
  } else {
    iov[0].iov_len = &proc->msg[proc->mlen] - proc->wr;
    iov[1].iov_len = proc->rd - proc->msg;
  }

  l = readv(proc->fd, iov, 2);
  if (l == 0) {
    if (proc->rd != proc->wr) send_output(proc, 1/*flush*/);
    FD_CLR(proc->fd, &watchedset);
    if (proc->pid == -1) join_child(proc);
    return;
  }
  if (l < 0) {
    if (errno == EAGAIN || errno == EWOULDBLOCK) return;
    xlog(LOG_WARNING, "%s: %s", proc->slot->pfix, strerror(errno));
    return;
  }
  proc->wr += l;
  if (proc->wr >= &proc->msg[proc->mlen])
    proc->wr -= proc->mlen;

  /* send back to client */
  send_output(proc, 0/*flush*/);

  /* update progress, if status reports are enabled */
  if (!*rbd_status) return;

  if (iov[0].iov_len < l) {
    iov[1].iov_len = l - iov[0].iov_len;
  } else {
    iov[0].iov_len = l;
    iov[1].iov_len = 0;
  }
  for(i=1; i>=0; i--) {
    for (t = iov[i].iov_base;
         (t = memrchr(iov[i].iov_base, '[', iov[i].iov_len));
         iov[i].iov_len = t - (char *)iov[i].iov_base) {
      for (p = 0, d = t+1; isdigit((int)*d); p = 10*p+(*d++-'0'));
      if (!(d[0] == '%' && d[1] == ']')) continue;

      if (proc->progress == p) return;

      proc->progress = p;
      write_status();
      return;
    }
  }
}


/* --- wait_children ------------------------------------------------------- */

static void
wait_children(void)
{
  struct rbd_group *group;
  struct rbd_slot *slot;
  struct rbd_proc *proc;
  pid_t p;
  int e;

  while ((p = wait4(-1, &e, WNOHANG, NULL)) > 0) {
    /* lookup child */
    proc = NULL;
    LIST_FOREACH(group, &groups, list) {
      LIST_FOREACH(slot, &group->slots, list) {
        if (slot->run && slot->run->pid == p) {
          proc = slot->run;
          break;
        }
        if (slot->kill && slot->kill->pid == p) {
          proc = slot->kill;
          break;
        }
      }
      if (proc) break;
    }
    if (!proc) {
      xlog(LOG_WARNING, "received status for old child %d", p);
      continue;
    }
    assert(proc->slot == slot);
    assert(proc->slot->group == group);

    proc->pid = -1;
    if (!proc->status) proc->status = e;

    if (!FD_ISSET(proc->fd, &watchedset))
      join_child(proc);
  }
}


/* --- do_event ------------------------------------------------------------ */

static void
do_event(void)
{
  struct rbd_clnt *clnt;
  fd_set set;
  int i, e;

  /* recv messages or signals */
  memcpy(&set, &watchedset, sizeof(set));
  expire = child = 0;
  e = pselect(maxwatched + 1, &set, NULL, NULL, NULL, &dfltmask);
  switch(e) {
    case -1: /* signals: children or grace delay expiration */
      if (errno != EINTR) xlog(LOG_ERR, "select: %s", strerror(errno));
      if (child) wait_children();
      if (expire) shuffle_spawn();
      break;

    case 0: /* cannot happen */
      xlog(LOG_ERR, "select timeout");
      break;

    default:
      if (child) wait_children();
      for (i=3; i<=maxwatched; i++) {
        if (!FD_ISSET(i, &set)) continue;

        /* child message */
        if (fdproc[i]) { recv_output(fdproc[i]); continue; }

        /* job request */
        if (fdclnt[i]) { recv_job(fdclnt[i]); continue; }

        /* connection request */
        clnt = xmalloc(sizeof(*clnt));
        clnt->fd = accept(i, NULL, &(socklen_t){0});
        if (clnt->fd < 0) {
          xlog(LOG_WARNING, "accept: %s", strerror(errno));
          free(clnt);
          continue;
        }
        fcntl(clnt->fd, F_SETFD, FD_CLOEXEC);
        clnt->lastout = NULL;
        LIST_INIT(&clnt->proc);

        fdwatch(clnt->fd);
        fdclnt[clnt->fd] = clnt;
      }
      if (expire) shuffle_spawn();
      break;
  }
}


/* --- save_state ---------------------------------------------------------- */

static char *
save_state(void)
{
  struct rbd_group *group;
  struct rbd_slot *slot;
  struct rbd_proc *w;
  char *list, name[256];
  size_t len;

  /* kill interruptible running jobs */
  LIST_FOREACH(group, &groups, list) LIST_FOREACH(slot, &group->slots, list) {
    if (slot->run && !slot->kill)
      spawn_job(slot, SPAWN_KILL);
  }

  /* save run queue */
  list = malloc(len = 64);
  *list = *"";
  LIST_FOREACH(group, &groups, list) {
    LIST_FOREACH(slot, &group->slots, list) {
      if (slot->run) {
        snprintf(name, sizeof name, " %s ", slot->run->name);
        if (!strstr(list, name)) {
          while (strlen(list) + strlen(name) + 1 >= len)
            list = xrealloc(list, len *= 2);
          strcat(list, name);
        }
      }
      LIST_FOREACH(w, &slot->wait, wlist) {
        snprintf(name, sizeof name, " %s ", w->name);
        if (!strstr(list, name)) {
          while (strlen(list) + strlen(name) + 1 >= len)
            list = xrealloc(list, len *= 2);
          strcat(list, name);
        }
      }
    }
  }

  if (*list) return list;
  free(list);
  return NULL;
}


/* --- main ---------------------------------------------------------------- */

int
main(int argc, char *const argv[])
{
  extern char *optarg;
  extern int optind;
  char *argv0 = argv[0];

  char dirs[MAXPATHLEN] = "";
  struct rbd_clnt *clnt;
  struct rbd_group *g;
  struct rbd_slot *s;
  struct rbd_proc *p;
  struct rbd_tgt *t;
  struct rbd_job *j;
  struct sigaction act;
  char *m;
  int i, c;

  /* parse command line options */
  while ((c = getopt(argc, argv, "hfc:b:l:g:s:d:o:")) != -1) {
    switch (c) {
      case 'f':
        daemonize = 0;
        break;

      case 'c':
        if (access(optarg, R_OK) || !realpath(optarg, config))
          xlog(LOG_ERR, "%s: %s", optarg, strerror(errno));
        break;

      case 'b':
        if (!realpath(optarg, bind_dir))
          xlog(LOG_ERR, "%s: %s", optarg, strerror(errno));
        break;

      case 's':
        strncpy(rbd_logstatus, optarg, sizeof(rbd_status)-sizeof("/%s"));
        strcat(rbd_logstatus, "/%s");
        snprintf(rbd_status, sizeof(rbd_status), rbd_logstatus,
                 "rbulk-dispatchd.html");
        strncpy(rbd_tmpstatus, rbd_status,
                sizeof(rbd_tmpstatus)-sizeof(".tmp"));
        strcat(rbd_tmpstatus, ".tmp");
        break;

      case 'l':
        run_user = optarg;
        break;

      case 'g':
        run_group = optarg;
        break;

      case 'd':
        grace = (int)strtol(optarg, &m, 0);
        if (*m || grace < 0 || grace > 3600)
          xlog(LOG_ERR, "%s: bad delay", optarg);
        break;

      case 'o':
        histlen = (int)strtol(optarg, &m, 0);
        if (*m || histlen < 0 || histlen > 1000)
          xlog(LOG_ERR, "%s: bad lenght", optarg);
        break;

      case 'h':
        usage(stdout, argv0);
        exit(0);

      case '?':
      default:
        usage(stderr, argv0);
        exit(2);
    }
  }
  if (argc > optind) {
    usage(stderr, argv0);
    exit(2);
  }

  /* random seed for shuffle_spawn() */
  srandom(time(NULL));

restart:
  /* load config */
  if (load_config())
    xlog(LOG_ERR, "cannot parse configuration in %s", config);
  if (!LIST_FIRST(&groups) || !LIST_FIRST(&jobs))
    xlog(LOG_ERR, "nothing to do in %s", config);

  /* load status data */
  load_statusdata(0/*unload*/);

  /* setup control socket */
  FD_ZERO(&watchedset);
  maxwatched = -1;
  fdproc = NULL;
  fdclnt = NULL;
  setup_socket(bind_dir);

  if (run_group) {
    struct group *grp;
    errno = 0;
    rgid = getgid();
    grp = getgrnam(run_group);
    if (!grp) {
      if (errno) xlog(LOG_ERR, "%s: %s", run_group, strerror(errno));
      xlog(LOG_ERR, "%s: unkown group", run_group);
    }
    if (chown(bind_addr, -1, grp->gr_gid))
      ;
    if (setegid(grp->gr_gid))
      xlog(LOG_ERR, "setgid: %s", strerror(errno));
    if (setgroups(0, NULL))
      xlog(LOG_ERR, "setgroups: %s", strerror(errno));
  }

  if (run_user) {
    struct passwd *pw;
    errno = 0;
    ruid = getuid();
    pw = getpwnam(run_user);
    if (!pw) {
      if (errno) xlog(LOG_ERR, "%s: %s", run_user, strerror(errno));
      xlog(LOG_ERR, "%s: unkown id", run_user);
    }
    if (chown(bind_addr, pw->pw_uid, -1))
      ;
    if (initgroups(pw->pw_name, pw->pw_gid))
      xlog(LOG_ERR, "initgroups: %s", strerror(errno));
    if (seteuid(pw->pw_uid))
      xlog(LOG_ERR, "seteuid: %s", strerror(errno));
  }

  if (*rbd_status) {
    snprintf(dirs, sizeof(dirs), rbd_logstatus, ".");
    if (mkdir(dirs, S_IRWXU | S_IRGRP|S_IXGRP | S_IROTH|S_IXOTH)) {
      if (errno != EEXIST)
        xlog(LOG_ERR, "%s: %s", dirs, strerror(errno));
    }
    snprintf(dirs, sizeof(dirs), rbd_logstatus, "rbulk-dispatchd");
    if (mkdir(dirs, S_IRWXU | S_IRGRP|S_IXGRP | S_IROTH|S_IXOTH)) {
      if (errno != EEXIST)
        xlog(LOG_ERR, "%s: %s", dirs, strerror(errno));
    }
  }


  /* daemonize */
  if (daemonize && !daemonized) {
    pid_t p;
    switch((p = fork())) {
      case -1:
        xlog(LOG_ERR, "cannot fork: %s", strerror(errno));

      case 0: /* child */ {
        int fd;

        if (setsid() == -1) exit(-1);
        if (chdir("/")) /*noop*/;

        fd = open("/dev/null", O_RDWR, 0);
        if (fd != -1) {
          dup2(fd, STDIN_FILENO);
          dup2(fd, STDOUT_FILENO);
          dup2(fd, STDERR_FILENO);
          if (fd > STDERR_FILENO) close(fd);
        }
        break;
      }

      default: /* parent */
        xlog(LOG_INFO, "daemonizing to pid %d", p);
        _exit(0);
    }

    daemonized = 1;
  }

  /* process signals in do_event() only */
  sigfillset(&act.sa_mask);
  if (sigprocmask(SIG_BLOCK, &act.sa_mask, &dfltmask))
    xlog(LOG_ERR, "sigprocmask: %s", strerror(errno));

  /* cleanup handlers */
  act.sa_sigaction = NULL;
  act.sa_handler = term_sighandler;
  sigfillset(&act.sa_mask);
  act.sa_flags = SA_RESTART;
  if (sigaction(SIGHUP, &act, NULL) ||
      sigaction(SIGINT, &act, NULL) ||
      sigaction(SIGQUIT, &act, NULL) ||
      sigaction(SIGTERM, &act, NULL))
    xlog(LOG_ERR, "sigaction: %s", strerror(errno));

  /* children handlers */
  act.sa_sigaction = NULL;
  act.sa_handler = child_sighandler;
  sigfillset(&act.sa_mask);
  act.sa_flags = SA_NOCLDSTOP|SA_RESTART;
  if (sigaction(SIGCHLD, &act, NULL) ||
      sigaction(SIGPIPE, &act, NULL))
    xlog(LOG_ERR, "sigaction: %s", strerror(errno));

  /* alarm handler */
  act.sa_sigaction = NULL;
  act.sa_handler = alarm_sighandler;
  sigfillset(&act.sa_mask);
  act.sa_flags = SA_RESTART;
  if (sigaction(SIGALRM, &act, NULL))
    xlog(LOG_ERR, "sigaction: %s", strerror(errno));

  /* process old jobs, enter event loop, then save */
  if (statedata) {
    clnt = xmalloc(sizeof(*clnt));
    clnt->fd = STDOUT_FILENO;
    clnt->lastout = NULL;
    LIST_INIT(&clnt->proc);
    for (m = strtok(statedata, " "); m; m = strtok(NULL, " "))
      sched_job(clnt, m);
    if (LIST_EMPTY(&clnt->proc)) free(clnt);
  }
  write_status();
  while(!leave) do_event();

  xlog(LOG_NOTICE, "cleaning up after signal %d", leave);
  m = save_state();

  /* cleanup */
  ngroups = nslots = 0;
  while((g = LIST_FIRST(&groups))) {
    while((s = LIST_FIRST(&g->slots))) {
      if (s->run) join_proc(s->run);
      if (s->kill) join_proc(s->kill);
      while((p = LIST_FIRST(&s->wait))) {
        LIST_REMOVE(p, wlist);
        join_proc(p);
      }
      while((p = LIST_FIRST(&s->hist))) {
        LIST_REMOVE(p, wlist);
        xfreeproc(p);
      }

      LIST_REMOVE(s, list);
      if (s->name) free((void *)s->name);
      if (s->desc) free((void *)s->desc);
      if (s->pfix) free(s->pfix);
      free(s);
    }
    LIST_REMOVE(g, list);
    if (g->name) free((void *)g->name);
    if (g->desc) free((void *)g->desc);
    free(g);
  }
  while((j = LIST_FIRST(&jobs))) {
    LIST_REMOVE(j, list);
    while((t = LIST_FIRST(&j->tgt))) {
      LIST_REMOVE(t, list);
      free(t);
    }
    free((void *)j->pattern);
    regfree(&j->reg);
    free((void *)j->desc);
    for(i = 0; i <= SPAWN_KILL; i++)
      free((void *)j->exec[i]);
    free(j);
  }
  if (fdproc) free(fdproc);
  if (fdclnt) free(fdclnt);

  write_status();
  load_statusdata(1/*unload*/);
  if (run_user) {
    if (seteuid(ruid))
      xlog(LOG_ERR, "setuid: %s", strerror(errno));
  }
  if (run_group) {
    if (setegid(rgid))
      xlog(LOG_ERR, "setgid: %s", strerror(errno));
  }
  unlink(bind_addr);
  if (m) {
    char name[256];
    FILE *f;
    snprintf(name, sizeof name, "%s/rbulk-dispatchd.save", bind_dir);
    f = fopen(name, "w");
    if (f) {
      fprintf(f, "%s\n", m);
      fclose(f);
      xlog(LOG_NOTICE, "saved state to %s", name);
    }
    free(m);
  }
  sigprocmask(SIG_SETMASK, &dfltmask, NULL);

  if (leave == SIGHUP) {
    xlog(LOG_NOTICE, "restarting on signal %d", leave);
    leave = 0;
    goto restart;
  }

  xlog(LOG_NOTICE, "terminating on signal %d", leave);
  return 0;
}
